/* luaglade.c  --  Lua interface to libglade
 *                                     (c)  Mark Colclough, 2006-2007
 * 
 */
 
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <dlfcn.h>
#include <ffi.h>
#include <gtk/gtk.h>
#include <glade/glade.h> 
#include <lauxlib.h>
#include "utils.h"
#include "helpers.h"
#include "lgtypes.h"
#include "version.h"
#include "notify.h"


int hash(const unsigned char *key); // defined in hashgtk.c
static void push_lua_widget(lua_State* L, GtkWidget* w); // defined below

void *dlhandle; // dl handle for dynamic access to gtk,atk,pango functions
void *gtkmainfunc; // so gtk_wrapper can know what it is calling


/* The hash() of the method name _is_ its encoded signature */
static unsigned short get_signature(const char *method_name) {
	unsigned short sig;
	sig = hash((const unsigned char *)method_name);
	return sig;
}

union gtk_types {
	void *p;
	long l;
	int i;
	unsigned int ui;
	double d;
	float f;
};

#define MAXARGS 4   // the this pointer plus 3 args from sig - or all if a function not a method 

/* Transfer params from lua and call the gtk function */
// called with 3 lightuserdata upvalues: gtk function pointer, sig(integer), gtk widget ptr
///!! watch out - the gtk_main call goes through here, but does not return until
///!! all the calls in the program have been processed.  Thus globals like sig and L will be
///!! bogus by this stage -- looks like those are the only ones to worry about
// the widget ptr is void*0 to indicate a function, not a method. 
static int gtk_wrapper(lua_State* L) {
	union gtk_types avalues[MAXARGS];  // C storage for arguments from Lua
	void *avalueptrs[MAXARGS];         // pointers to arguments
	union gtk_types atempvalues[MAXARGS]; // C storage for pointer targets
	ffi_type *atypes[MAXARGS];         // ffi type pointers for arguments
	unsigned int nargs;
	union gtk_types rvalue;    // return value
	ffi_cif cif;          // call interface structure for ffi
	void *fptr;      // gtk function pointer
	unsigned short sig;   // signature of gtk function
	unsigned int code; // one item from the sig
	int i, ret;
	void *wptr;   // widget pointer or nul if a non-widget function is being called
	
	for (i=0; i<MAXARGS; i++) {
		avalueptrs[i] = &avalues[i];
	}
	
	fptr = lua_touserdata(L, lua_upvalueindex(1) );           // function ptr
	sig = (unsigned short)(unsigned int)lua_touserdata(L, lua_upvalueindex(2) );  // function sig
	wptr = lua_touserdata(L, lua_upvalueindex(3) );
	
	if ((sig & ERRFLAG) == ERRFLAG) {
		switch (sig) {
			case ERRNARGS: warn("function needs more args than are supported\n"); break;
			case ERRTYPE:  warn("function has unsupported arg types\n"); break;
			default:       warn("function has mystery error signature 0x%x\n", sig); break;
		}
		return 0;
	}
	
	code = sig & 0x000F;
	dbg("ret type code %d\n", code);
	ffi_type *rtype = &ffi_type_void;   // ffi type pointer for return value
	switch(code) {
		case T_VOID:
			dbg("ret type is void\n");
			rtype = &ffi_type_void; break;
		case T_BOOL:
			dbg("ret type is sint (for boolean)\n");
			rtype = &ffi_type_sint; break;
		case T_INT:
			dbg("ret type is sint (for gint)\n");
			rtype = &ffi_type_sint; break;
		case T_UINT:
			dbg("ret type is uint (for guint)\n");
			rtype = &ffi_type_uint; break;
		case T_GUINTP:
			dbg("ret type is ptr (for guint*)\n");
			rtype = &ffi_type_pointer; break;
		case T_FLOAT:
			dbg("ret type is float (for gfloat)\n");
			rtype = &ffi_type_float; break;
		case T_DOUBLE:
			dbg("ret type is double (for gdouble)\n");
			rtype = &ffi_type_double; break;
		case T_STRING:
			dbg("ret type is ptr (for const (g)char*)\n");
			rtype = &ffi_type_pointer; break;
		case T_WIDGET:
			dbg("ret type is ptr (for *GtkWidget)\n");
			rtype = &ffi_type_pointer; break;
	}
	
	// Keeping these in sync is hard!
	int sp = 1;  // pointer to arguments on lua stack
	int np = 1;  // pointer to nibble of sig  (sig >> (np*4)) & 0x000F
	int gp = 0;  // pointer to gtk arguments in ffi avalues and atypes
	
	// if it is a method call, first arg comes as upvalue, not on stack
	if (wptr) {
		dbg("providing widget pointer %p as first argument\n", wptr);
		avalues[gp].p = wptr;
		atypes[gp] = &ffi_type_pointer;
		gp++;
	}
	
	while(gp < MAXARGS) {
		code = (sig >> (np*4)) & 0x000F;
		if (code == T_VOID) {
			dbg("arg %d is void - all done\n", i);
			break;
		}
		dbg("copying stack arg %d, having code %d\n", sp, code);
		switch(code) {
			case T_BOOL:
				avalues[gp].i = luaL_checkint(L, sp);
				atypes[gp] = &ffi_type_sint;
				dbg("copying int (bool) value %d\n", avalues[gp].i); 
				break;
			case T_INT:
				avalues[gp].i = luaL_checkint(L, sp);
				atypes[gp] = &ffi_type_sint;
				dbg("copying int (int) value %d\n", avalues[gp].i);
				break;
			case T_UINT:
				avalues[gp].ui = luaL_checkint(L, sp);
				atypes[gp] = &ffi_type_uint;
				dbg("copying unsigned int value %d\n", avalues[gp].ui);
				break;
			case T_GUINTP:
				atempvalues[gp].ui = luaL_checkint(L, sp);
				avalues[gp].p = &atempvalues[gp].ui;
				atypes[gp] = &ffi_type_pointer;
				dbg("copying pointer (to uint) pointing at %d\n", atempvalues[gp].ui); 
				break;
			case T_FLOAT:
				avalues[gp].f = (float)luaL_checknumber(L, sp);
				atypes[gp] = &ffi_type_float;
				dbg("copying float value %f\n", avalues[gp].f); 
				break;
			case T_DOUBLE:
				avalues[gp].d = luaL_checknumber(L, sp);
				atypes[gp] = &ffi_type_double;
				dbg("copying double value %lf\n", avalues[gp].d);
				break;
			case T_STRING:
				avalues[gp].p = (void*)luaL_checkstring(L, sp);
				atypes[gp] = &ffi_type_pointer;
				dbg("copying pointer to string, pointing at %s\n",(char*)avalues[gp].p);
				break;
			case T_WIDGET:
				lua_getfield(L, sp, "address");  // the widget table   // ... wt
				avalues[gp].p =(void*)lua_touserdata(L, -1);
				lua_pop(L,1);                                          // ...
				atypes[gp] = &ffi_type_pointer;
				dbg("copying pointer to widget %p\n", avalues[gp].p);
				break;
		}
		gp++; sp++; np++;		
	} // end of loop over arguments
	
	
	nargs = gp;  // get this wrong and segfault city awaits
	dbg("nargs for ffi is %d\n", nargs);
	
	// call the function
	ret = ffi_prep_cif(&cif, FFI_DEFAULT_ABI, nargs, rtype, atypes);	
	
	if (ret == FFI_OK)
		dbg("cif creation result OK: %d\n", ret);
	else
		dbg("cif creation result FAILED: %d\n", ret);
	
	ffi_call(&cif, fptr, &rvalue, avalueptrs);
	dbg("ffi_call done\n");
	
	// extract the results
	//  ... first from the return value
	int nreturned = 0;
	switch(sig & 0x000F) {
		case T_VOID:
			dbg("return type void, returning no value\n");
			break;
		case T_BOOL:
			dbg("returning bool value %d\n", rvalue.i);
			lua_pushboolean(L, rvalue.i);
			nreturned++;
			break;
		case T_INT:
			dbg("returning int value %d\n", rvalue.i);
			lua_pushnumber(L, (double)rvalue.i);
			nreturned++;
			break;
		case T_UINT:
			dbg("returning uint value %d\n", rvalue.ui);
			lua_pushnumber(L, (double)rvalue.ui);
			nreturned++;
			break;
		case T_GUINTP:
			dbg("returning int from uintp, value %d\n", *(unsigned*)rvalue.p);
			lua_pushnumber(L, (double)*(unsigned*)rvalue.p);
			nreturned++;
			break;
		case T_FLOAT:
			dbg("returning float value %f\n", rvalue.f);
			lua_pushnumber(L, (double)rvalue.f);
			nreturned++;
			break;
		case T_DOUBLE:
			dbg("returning double value %lf\n", rvalue.d);
			lua_pushnumber(L, rvalue.d);
			nreturned++;
			break;
		case T_STRING:
			dbg("returning string value %s\n", (char*)rvalue.p);
			lua_pushstring(L, (char*)rvalue.p);
			nreturned++;
			break;
		case T_WIDGET:
			dbg("returning widget with addr %p\n", rvalue.p);
			push_lua_widget(L, (GtkWidget*)rvalue.p);
			nreturned++;
			break;
	}
	// ... and later from any non-const ptrs in arg list -- ignore for now.
	// finally will push these too and increment return value

	return nreturned;
}

// One of these is made in the makegui() call and is used for all autoconnected
// signals. More could be made if a separate call fromm Lua wanted to connect
// more callbacks
struct guicontext {
	int gui_table_ref;
	lua_State *L;
};

// Number and types of parameters that the signal expects to deliver to its handler,
// plus the lua handler and the user data that was appended to the handler name in
// glade (or nil if none was).  We malloc one of these for each connection.
struct cbinfo {
	GSignalQuery signal_info;
	int handler_ref;
	int udata_ref;
	struct guicontext *gc;
};

// This is where GTK signals are connected (via the correct cb_despatchN).
// This function calls the user's Lua handler 
static void cb_despatch(void *widget, struct cbinfo *cb_params, int nsigp, void* sig_params[]) {
	dbg("despatch n=%d\n", nsigp);
	int i;
	lua_State *L = cb_params->gc->L;
	int tos = lua_gettop(L);
	
	lua_rawgeti(L, LUA_REGISTRYINDEX, cb_params->handler_ref);   // f (the lua handler)
	
	// next, the lua widget table w, given the address of the gtk widget
	lua_rawgeti(L, LUA_REGISTRYINDEX, cb_params->gc->gui_table_ref); // f g
	lua_pushlightuserdata(L, widget);     // f g *w
	lua_gettable(L, -2);                  // f g w    // forces lazy lookup if needed
	lua_remove(L, -2);                    // f w
	
	// then the user data set in glade (or nil) 
	lua_rawgeti(L, LUA_REGISTRYINDEX, cb_params->udata_ref);     // f w u
	
	// then the name of the signal
	lua_pushstring(L, cb_params->signal_info.signal_name);  // f w u s
	
	// then any signal parameters
	for (i=0; i<nsigp; i++) {
		lua_pushnumber(L, (int)sig_params[i]);  // f w u s pi
	}
	// finally call lua
	lua_call(L, 3+nsigp, 0);
	
	//restore stack
	lua_settop(L, tos);
}

static void cb_despatch0(void *widget, struct cbinfo *cb_params) {  //strip this one 
	cb_despatch(widget, cb_params, 0, 0);
}

static void cb_despatch1(void *widget, void* sig_param1, struct cbinfo *cb_params) {
	void *sig_params[1];
	sig_params[0] = sig_param1;
	cb_despatch(widget, cb_params, 1, sig_params);
}

static void cb_despatch2(void *widget, void* sig_param1, void* sig_param2,
                         struct cbinfo *cb_params) {
	void *sig_params[2];
	sig_params[0] = sig_param1; sig_params[1] = sig_param2;
	cb_despatch(widget, cb_params, 2, sig_params);	
}

static void cb_despatch3(void *widget, void* sig_param1, void* sig_param2,
                         void* sig_param3, struct cbinfo *cb_params) {
	void *sig_params[3];
	sig_params[0] = sig_param1; sig_params[1] = sig_param2;  sig_params[2] = sig_param3;
	cb_despatch(widget, cb_params, 3, sig_params);	
}

static void *cb_despatchers[] = { cb_despatch0, cb_despatch1, cb_despatch2, cb_despatch3};


// Called by the glade autoconnect, runs once per connection.
// We expect userdata to be tacked on to the handler_name with a #, to evade the lack
// of an arbitrary userdata facility in glade -- the udata field just looks up a second
// widget, and does not do what one might hope.                                    
static void signal_connector(const gchar *decorated_handler_name,
                             GObject *object,  // widget emitting signal
                             const gchar *signal_name,
                             const gchar * UNUSED(signal_data),
                             GObject * UNUSED(connect_object),
                             gboolean after,
                             gpointer gc) {  // from autoconnect
	dbg("in connector: handler=%s signal=%s  obj=%p  gc=%p\n",
	        decorated_handler_name, signal_name, object, gc);

	char handler_name_copy[255];
	char *handler;
	char *udata;
	char *table;
	char *dot;
	
	struct lua_State* L = ((struct guicontext*)gc)->L;
	
	struct cbinfo *cb_params;
	cb_params = (struct cbinfo*)malloc(sizeof (struct cbinfo));
	
	// Save the provided guicontext for callback time  
	cb_params->gc = gc;
	
	// Extract lua handler name and user data from decorated handler name.
	strncpy(handler_name_copy, decorated_handler_name, sizeof handler_name_copy);  // r/w copy
	handler = handler_name_copy;
	if ((udata = strchr(handler, '#'))) {
		*udata = 0;
		udata++;
	}

	dbg("parsed lua_handler=%s.  udata=%s.\n", handler, udata);
	
	// Resolve handler function and save for callback time. If name has a dot,
	// treat as table.handler 
	if ((dot = strchr(handler, '.'))) {
		table = handler;
		*dot = 0;
		handler = ++dot;
		dbg("Looking in table %s for handler function %s\n", table, handler);
		lua_getfield(L, LUA_GLOBALSINDEX, table);                  // t
		if(!lua_istable(L, -1)) {
			warn("no table %s to search for handler %s\n", table, handler);
			lua_pop(L, 1);              //  {empty}
			goto err;
		}
		// No point in checking whether function is there, because dl lazy lookup
		// will bail before the getfield returns.
		lua_getfield(L, -1, handler);                              // t h
		lua_remove(L, -2);                                         // h
	} else {
		dbg("looking in global table for handler function %s\n", handler);
		lua_getfield(L, LUA_GLOBALSINDEX, handler);                // h
	}
	if (!lua_isfunction(L, -1)) {
		lua_pop(L, 1);                                               // {empty}
		lua_getfield(L, LUA_GLOBALSINDEX, "default_handler");         // h'
		if (!lua_isfunction(L, -1)) {
			warn("Lua handler %s not found and there is no default_handler\n", handler);
			lua_pop(L, 1);                               // {empty}
			goto err;
		}
	}                                                          // (h or h')
	cb_params->handler_ref = luaL_ref(L, LUA_REGISTRYINDEX);   //  {}  {REG.(c->h)=luahandler}
	
	// save the udata for callback time
	(udata) ? lua_pushstring(L, udata) : lua_pushnil(L);       // (udata or nil)
	cb_params->udata_ref = luaL_ref(L, LUA_REGISTRYINDEX); // {}    {REG.(c->u)=udata}
	
	// Find out what parameters come with this signal
	guint signal_id = g_signal_lookup(signal_name, G_OBJECT_TYPE(object));
	g_signal_query(signal_id, &cb_params->signal_info);
	if (!signal_id || cb_params->signal_info.signal_id != signal_id) {
		warn("signal %s from object %p has id %d and is not known\n", signal_name, object, signal_id);
		goto err;
	}

	// Connect to despatcher that matches number of parameters
	guint nparams = cb_params->signal_info.n_params;
	if (nparams > 3) {
		warn("signal %s sent to %s has unsupported number of parameters: %d\n",
		       signal_name, decorated_handler_name, nparams);
		goto err;
	}
	dbg("nparams=%d\n", nparams);
	dbg("after=%d , ", after);
	int conid;
	if (after) {
		conid = g_signal_connect_after(object, signal_name, G_CALLBACK(cb_despatchers[nparams]), cb_params);
		dbg("after=yes\n");
	} else {
		conid = g_signal_connect(object, signal_name, G_CALLBACK(cb_despatchers[nparams]), cb_params);
		dbg("after=no\n");
	}
	return;
err:
	free(cb_params);
	return;
}

///////////  Background handling : timeouts and idle processes   ////////////////
struct bginfo {
	lua_State *L;
	int handler_ref;
	int udata_ref;
	int tag;
};

static gboolean bg_despatch(struct bginfo *bg) {
	///dbg("bg called with bg %p\n", bg);
	lua_State *L = bg->L;
	int tos = lua_gettop(L);  // preserve stack level for exit
	lua_rawgeti(L, LUA_REGISTRYINDEX, bg->handler_ref);   // f
	lua_rawgeti(L, LUA_REGISTRYINDEX, bg->udata_ref);     // f u
	lua_call(L, 1, 1);                                    // again
	int again = lua_toboolean(L, -1);                     // again
	lua_settop(L, tos);
	return again;
}
	
static int l_addtimeout(lua_State* L) {      // interval luafunc (udata?)
	int interval = luaL_checkint(L, 1);
	luaL_checktype(L, 2, LUA_TFUNCTION);
	lua_settop(L, 3);                         // interval luafunc udata/nil
	struct bginfo *bg = (struct bginfo*)malloc(sizeof(struct bginfo));
	bg->L = L;
	bg->udata_ref = luaL_ref(L, LUA_REGISTRYINDEX);   // interval luafunc  {REG.(b->u)=udata}
	bg->handler_ref = luaL_ref(L, LUA_REGISTRYINDEX); // interval   {REG.(b->h)=luafunc}
	bg->tag = g_timeout_add(interval, (gboolean(*)(gpointer))bg_despatch, bg);
	lua_pushlightuserdata(L, bg);                     // interval *bg
	dbg("timeout set with bg %p\n", bg);
	return 1;            // {*bg}
}

static int l_addidle(lua_State* L) {         // luafunc (udata?)
	luaL_checktype(L, 1, LUA_TFUNCTION);
	lua_settop(L, 2);                                           // luafunc udata/nil 
	struct bginfo *bg = (struct bginfo*)malloc(sizeof(struct bginfo));
	bg->L = L;
	bg->udata_ref = luaL_ref(L, LUA_REGISTRYINDEX);   // luafunc  {REG.(b->u)=udata}
	bg->handler_ref = luaL_ref(L, LUA_REGISTRYINDEX); // {REG.(b->h)=luafunc}
	bg->tag = g_idle_add( (gboolean(*)(gpointer))bg_despatch, bg);
	lua_pushlightuserdata(L, bg);                     // *bg
	dbg("timeout set with bg %p\n", bg);
	return 1;            // {*bg}	
}

static int l_removebg(lua_State* L) {   // *bg
	struct bginfo *bg = (struct bginfo*) lua_touserdata(L, 1);
	if (!bg) { lua_pushboolean(L, 0 ); return 1; }
	int ok = g_source_remove(bg->tag);
	if(!ok) { lua_pushboolean(L, 0 ); return 1; }
	lua_pushboolean(L, 1 );   // *bg ok	
	luaL_unref(L, LUA_REGISTRYINDEX, bg->udata_ref);
	luaL_unref(L, LUA_REGISTRYINDEX, bg->handler_ref);
	free(bg);
	return 1;
}

//////////////////////////////////////////////////////////////////////////////////


/* resolve function name for given widget and stash for next time */
static int l_getwidgetmethod(lua_State* L) {
	/*
	const char* mshortname = lua_tostring(L, -1);
	lua_getfield(L, 1, "address");
	void* waddress = lua_touserdata(L, -1);
	lua_pop(L,1);
	dbg("gwm for obj at addr %p\n", waddress);
	*/
	
	const char* mshortname = lua_tostring(L, 2);
	lua_getfield(L, 1, "address");
	void* waddress = lua_touserdata(L, -1);
	
	
	void *funcptr = NULL;
	char method_name[256];
	int otype = GTK_OBJECT_TYPE(waddress);
	dbg("resolving %s for widget type %s at %p\n",mshortname, g_type_name(otype), waddress);
	
	// Construct function names of the form gtk_font_button_set_text from gtk type
	// names like GtkFontButton and short method names like set_text.  Try all
	// the types from which GtkFontButton inherits until a suitable function is
	// found in the library.	/// Hunt down a real method
	while (otype != 0 && !funcptr) {
		char *p = method_name;
		const char *q = g_type_name(otype);
		*p++ = tolower(*q++);  // lowercase first char,  without leading _
		// lowercase the rest, inserting _ before (lowered) uppercase
		// a use for the comma operator...
		while( ((isupper(*q)) && (*p++ = '_')), (*p++ = tolower(*q++) ) );
		strcat(method_name, "_");
		strcat(method_name, mshortname);
		funcptr = dlsym(dlhandle, method_name);
		dbg("trying %s %p\n", method_name, funcptr);
		otype = g_type_parent(otype);
	}
	
	if (funcptr) {
		// look up signature for this method
		int sig = get_signature(method_name);
		if (sig < 0) {   /// can it ever fail?? 
			warn("no signature for method %s\n", method_name);
			return 0;
		}
		
		lua_pushlightuserdata(L, funcptr);
		lua_pushlightuserdata(L, (void*)sig);
		lua_pushlightuserdata(L, waddress);
		lua_pushcclosure(L, &gtk_wrapper, 3);
		
		lua_pushvalue(L, -1);
		lua_setfield(L, 1, mshortname);
		return 1;

	} else {
		// never really returns
		return luaL_error(L, "method %s is not in library\n", mshortname);
	}
}

// meta method for inserting global gtk functions in global module table
static int lazy_glob(lua_State *L) {               // t fname
	void *funcptr = NULL;
	const char *funcname = lua_tostring(L, -1);
	dbg("looking up %s in library\n", funcname);
	funcptr = dlsym(dlhandle, funcname);
	if (!funcptr) {
		return luaL_error(L, "function %s is not in library\n", funcname); //never returns
	}
	int sig = get_signature(funcname);  // may be bogus but will never fail
	dbg("function %s found at %p with sig 0x%x\n", funcname, funcptr, sig);
	lua_pushlightuserdata(L, funcptr);          // t fname fptr
	lua_pushlightuserdata(L, (void*)sig);       // t fname fptr fsig
	lua_pushlightuserdata(L, (void*)0);         // t fname fptr fsig wptr  widget ptr=0 for non-method
	lua_pushcclosure(L, &gtk_wrapper, 3);	     // t fname fclosure  -- is gtkwrapper OK? prob if we fix wptr
	lua_pushvalue(L, -1);                       // t fname fclosure fclosure
	lua_insert(L, 2);                           // t fclosure fname fclosure
	lua_settable(L, 1);                         // t fclosure
	return 1;
}

// Accept a GtkWidget* and push a lua widget referring to it. 
static void push_lua_widget(lua_State* L, GtkWidget* w) {
	const char *object_type = g_type_name(GTK_OBJECT_TYPE(w));
	dbg("Luifying a widget of type %s\n", object_type);
	
	// Make a widget table t for the widget
	lua_newtable(L);                         // t 
	
	lua_pushlightuserdata(L, w);             // t *w
	lua_setfield(L, -2, "address");          // t  {t.address= *w}
	
	// Metamethod to lazily add widget methods to this table
	lua_pushcfunction(L, l_getwidgetmethod); // t getwm
	lua_setfield(L, -2, "__index");          // t  {t.__index=getwm}
	
	// Add special functions for GtkTreeView, binding the widget table
	// as an upvalue, in order to pass the model amongst themselves.
	if (!strcmp(object_type, "GtkTreeView")) {         // t
		lua_pushvalue(L, -1);                           // t t 
		lua_pushcclosure(L, &tree_view_set_columns, 1); // t tvsc
		lua_setfield(L, -2, "set_columns");             // t  {t.s_c=tvsc}
		
		lua_pushvalue(L, -1);                           // etc...
		lua_pushcclosure(L, &tree_view_dump_data, 1);
		lua_setfield(L, -2, "dump_data");
		
		lua_pushvalue(L, -1);
		lua_pushcclosure(L, &tree_view_set_data, 1);
		lua_setfield(L, -2, "set_data");
	}                                                  // t
	
	// Recurse! - add a buffer widget to any TextView widget
	if (!strcmp(object_type, "GtkTextView")) {         // t
		GtkTextBuffer *buff = gtk_text_view_get_buffer((GtkTextView*)w);
		push_lua_widget(L, (GtkWidget*)buff);           // t buft
		lua_setfield(L, -2, "buffer");                  // t {buffer=buft}
		
		lua_pushlightuserdata(L, w);  // t *w
		lua_pushlightuserdata(L, buff);  // t *w *b
		lua_pushcclosure(L, l_text_view_add, 2);  // t closure_add
		lua_setfield(L, -2, "add");          // t {"add"=closure_add}
	}
	
	if (!strcmp(object_type, "GtkTextBuffer")) {        // t
		// add helper method
		lua_pushlightuserdata(L, w);  // t *w
		lua_pushcclosure(L, l_text_view_insert_at_line, 1);  // t closure_i_a_l
		lua_setfield(L, -2, "insert_at_line");          // t {"ial"=closure_i_a_l}
	}
	
	// make table its own metatable
	lua_pushvalue(L,-1);                               // t t'
	lua_setmetatable(L,-2);                            // t  {t.meta=t'}
}

/* Look up widget by name or by address in glade xml, return table holding functions
 * and other per-widget data.  This is the __index metamethod for g
 * Stash the returned table for next time in the top-level gui table */
static int l_getwidget(lua_State* L) {      // g wname  or  g *w
	const char* wname = NULL;
	GtkWidget* w = NULL;
	GladeXML *xml; 

	if (lua_isstring(L, -1)) {
		lua_getfield(L, -2, "xml");           // g wname *xml
		xml = lua_touserdata(L, -1);          // g wname *xml
		lua_pop(L, 1);                        // g wname
		wname = lua_tostring(L, -1);
		w = glade_xml_get_widget (xml, wname);
	} else if (lua_islightuserdata(L, -1)){
		w = lua_touserdata(L, -1);
		wname = glade_get_widget_name(w);
	}
	if (!w) return 0;  // nothing returned or saved if widget name not found.
	
	lua_pop(L, 1);                                     // g
	push_lua_widget(L, w);                             // g t 
	// save the new widget table t by widget name in the gui table g for next time 
	lua_pushvalue(L, -1);                              // g t t
	lua_setfield(L, 1, wname);                         // g t  {g.wname = t}
	
	// also save it by widget address
	lua_pushlightuserdata(L, w);                       // g t *w
	lua_pushvalue(L, -2);                              // g t *w t
	lua_settable(L, 1);                                // g t  {g.*w=t}
	
	return 1;                    // {t}
}


/*     makegui 
 * Construct widgets from a comment in the Lua source, a glade xml string or a 
 * separate glade file.  In the first case, the glade xml must be in a block comment:
 * --[===[<glade-interface>...</glade-interface>]===] and the call is makegui("===")
 * where the number of equals signs must correspond to those in the comment.
 * Alternatively, provide the xml string (which must begin with "<") as the first
 * argument.  As a final alternative, give the name of a glade xml file as the first
 * argument.  The root node and translation domain are optional second and third
 * arguments.
 * This function returns a table ("the gui") which initially just has an __index
 * metamethod, for accessing widgets by name.  The widgets (which are represented as
 * tables) are added to the gui table by name, whenever __index locates them, for
 * fast subsequent access.
 */
static int l_makegui(lua_State *L) {
	const char *xml_string_or_file;
	const char *source_file_name = NULL;
	size_t xml_length;
	const char *xml_root = NULL;
	const char *xml_domain = NULL;
	GladeXML *xml; // the gtk representation we hope to create 
	int nargs;
	luaL_Buffer xml_buffer;
	char *bufp;
	int xml_fd;
	ssize_t nread;
	static int gtk_initialised = 0;
	// gc is sent on to the autoconnect, where it is shared by all connections.
	struct guicontext *gc = (struct guicontext*)malloc(sizeof(struct guicontext));
	gc->L = L;
	
	if (! gtk_initialised) {
		dlhandle = dlopen ("gtk-x11-2.0", RTLD_LAZY);
		gtk_init(NULL,NULL);
		gtk_initialised = 1;
	}
	
	nargs = lua_gettop(L);
	xml_string_or_file = luaL_checklstring(L, 1, &xml_length);
	if (nargs > 1) {
		xml_root = luaL_checkstring(L, 2);
	}
	if (nargs > 2) {
		xml_domain = luaL_checkstring(L, 3);
	}
	
	switch(*xml_string_or_file) {
		case '=':    // extract xml from lua file, then hand over the buffer
			lua_getglobal (L, "arg");  // get arg[0] - note we leave arg{} and arg[0] on stack
			lua_pushinteger (L, 0);
			lua_gettable (L, -2);
			source_file_name = lua_tostring (L, -1);
			// now read in the whole file and parse out the xml part using string.match()
			xml_fd = open(source_file_name, O_RDONLY);
			if (xml_fd < 0) {
				return luaL_error(L, "Failed to open xml-in-lua file %s", source_file_name);
			}
			luaL_buffinit(L, &xml_buffer);
			bufp = luaL_prepbuffer(&xml_buffer);  // of size LUAL_BUFFERSIZE
			while( (nread = read(xml_fd, bufp, LUAL_BUFFERSIZE)) > 0 ) {
				luaL_addsize(&xml_buffer, nread);
				bufp = luaL_prepbuffer(&xml_buffer);
			}
			close(xml_fd);
			if (nread < 0) {
				luaL_error(L, "Failed to read xml-in-lua file %s", source_file_name);
			}
			luaL_pushresult(&xml_buffer);  // whole source file is now in string at tos
			lua_getglobal(L, "string");
			lua_getfield(L, -1, "match");
			lua_pushvalue(L, -3); // copy of xml_buffer as first arg for string.match
			lua_pushfstring(L, "%%-%%-%%[%s%%[(.*)%%]%s%%]", xml_string_or_file, xml_string_or_file);
			lua_call(L, 2, 1);
			xml_string_or_file = lua_tolstring(L, -1, &xml_length);  // now we have extracted the xml part
			xml = glade_xml_new_from_buffer(xml_string_or_file, xml_length, xml_root, xml_domain);
			if (!xml) {
				return luaL_error(L, "Failed to extract good xml from lua at %s", source_file_name);
			}
			break;
		case '<':    // We are given the XML buffer
			xml = glade_xml_new_from_buffer(xml_string_or_file, xml_length, xml_root, xml_domain);
			if (!xml) {
				return luaL_error(L, "Failed to parse glade string");
			}
			break;
		default:        // We are given the xml file name
			xml = glade_xml_new(xml_string_or_file, xml_root, xml_domain);
			if (!xml) {
				return luaL_error(L, "Failed to read glade file %s", xml_string_or_file);
			}
			break;
	}

	lua_newtable(L);                                       // g
	lua_pushlightuserdata(L, xml);                         // g *xml	
	lua_setfield(L, -2, "xml");                            // g  {g.xml=xml}
	lua_pushcfunction(L, l_getwidget);                     // g l_getwidget
	lua_setfield(L, -2, "__index");                        // g  {g.__index=l_getwidget}
		
	lua_pushvalue(L,-1);  // table is its own metatable    // g g
	lua_setmetatable(L,-2);                                // g  {g.meta=g}
	// make this table ("the gui table") widely accessable from C, via gc. 
	lua_pushvalue(L,-1);                                   // g g
	gc->gui_table_ref = luaL_ref(L, LUA_REGISTRYINDEX);    // g  {REG.g_t_r=g}

	glade_xml_signal_autoconnect_full(xml, signal_connector, (void*)gc );
	return 1;
}

// Functions to be directly available in the module namespace ("G")
static const struct luaL_Reg luaglade [] = {
	{"make_gui", l_makegui},
	{"add_timeout", l_addtimeout},
	{"remove_timeout", l_removebg},
	{"add_idle", l_addidle},
	{"remove_idle", l_removebg},
	{"add_watch", l_addwatch},
	{"remove_watch", l_removewatch},
	{"set_cursor", l_set_cursor},
	{"do_events", l_do_events},
	{"isset", l_isset},    /// should really be in lualib, not here
	{NULL, NULL}
};

int luaopen_luaglade(lua_State *L) {
	lua_newtable(L);
	lua_replace(L, LUA_ENVIRONINDEX);      // environment table (used by inotify)
	luaL_register(L, "G", luaglade);       // luaglade
	// add other stuff to the module table
	lua_pushstring(L, VERSION);               // luaglade version
	lua_setfield(L, -2, "version");           // luaglade  {luaglade.version=versionstr}
	
	// make entries for some selected enum constants
	set_cursor_types(L);         // GDK cursor types (just some popular ones)
	set_notify_masks(L);         // inotify event mask constants
	
	// add __index metamethod and make table its own metatable 
	lua_pushcfunction(L, lazy_glob);          // luaglade lazy_glob
	lua_setfield(L, -2, "__index");           // luaglade  {luaglade.__index=lazy_glob}
	lua_pushvalue(L,-1);                      // luaglade luaglade'
	lua_setmetatable(L,-2);                   // luaglade  {luaglade.meta=luaglade'}
	return 1;
}
